VERSION 5.00
Begin VB.Form Form1 
   Caption         =   "howto_apollonian_gasket"
   ClientHeight    =   4935
   ClientLeft      =   60
   ClientTop       =   450
   ClientWidth     =   4935
   LinkTopic       =   "Form1"
   ScaleHeight     =   4935
   ScaleWidth      =   4935
   StartUpPosition =   3  'Windows Default
   Begin VB.PictureBox picPacking 
      AutoRedraw      =   -1  'True
      Height          =   4695
      Left            =   120
      ScaleHeight     =   4635
      ScaleWidth      =   4635
      TabIndex        =   0
      Top             =   120
      Width           =   4695
   End
End
Attribute VB_Name = "Form1"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = False
Attribute VB_PredeclaredId = True
Attribute VB_Exposed = False
Option Explicit

Private Sub Form_Load()
    picPacking.BackColor = vbYellow
    picPacking.ScaleMode = vbPixels

    MakeImage
End Sub
Private Sub Form_Resize()
    MakeImage
End Sub

Private Sub MakeImage()
Dim width As Integer

    If (picPacking.ScaleWidth < picPacking.ScaleHeight) Then
        width = picPacking.ScaleWidth
    Else
        width = picPacking.ScaleHeight
    End If

    FindApollonianPacking picPacking, width
End Sub

' Find the Apollonian packing.
Private Sub FindApollonianPacking(ByVal pic As PictureBox, ByVal width As Integer)
Dim Radius As Single
Dim X As Single
Dim gasket_height As Single
Dim Y As Single
Dim circle0 As MyCircle
Dim circle1 As MyCircle
Dim circle2 As MyCircle
Dim big_circle As MyCircle
Dim level As Integer

    pic.Cls

    ' Create the three central tangent circles.
    Radius = width * 0.225
    X = width \ 2
    gasket_height = 2 * CSng(Radius + 2 * Radius / Sqr(3))
    Y = (width - gasket_height) / 2 + Radius
    Set circle0 = New MyCircle
    circle0.X = X
    circle0.Y = Y
    circle0.Radius = Radius

    X = X - Radius
    Y = Y + CSng(Radius * Sqr(3))
    Set circle1 = New MyCircle
    circle1.X = X
    circle1.Y = Y
    circle1.Radius = Radius
    X = X + 2 * Radius
    Set circle2 = New MyCircle
    circle2.X = X
    circle2.Y = Y
    circle2.Radius = Radius

    ' Draw the three central circles.
    circle0.Draw pic, vbRed
    circle1.Draw pic, vbRed
    circle2.Draw pic, vbRed

    ' Find the circle that contains them all.
    Set big_circle = FindApollonianCircle( _
        circle0, circle1, circle2, -1, -1, -1)
    big_circle.Draw pic, vbRed

    ' Set level to smaller values such as 3 to see partially drawn gaskets.
    level = 10000

    ' Find the central circle.
    FindCircleOutsideAll level, pic, circle0, circle1, circle2

    ' Find circles tangent to the big circle.
    FindCircleOutsideTwo level, pic, circle0, circle1, big_circle
    FindCircleOutsideTwo level, pic, circle1, circle2, big_circle
    FindCircleOutsideTwo level, pic, circle2, circle0, big_circle
End Sub

' Draw a circle tangent to these three circles and that is outside all three.
Private Sub FindCircleOutsideAll(ByVal level As Integer, ByVal pic As PictureBox, ByVal circle0 As MyCircle, ByVal circle1 As MyCircle, ByVal circle2 As MyCircle)
Dim new_circle As MyCircle

    Set new_circle = FindApollonianCircle(circle0, circle1, circle2, 1, 1, 1)
    If (new_circle Is Nothing) Then Exit Sub
    If (new_circle.Radius < 0.1) Then Exit Sub

    new_circle.Draw pic, vbRed

    level = level - 1
    If (level > 0) Then
        FindCircleOutsideAll level, pic, circle0, circle1, new_circle
        FindCircleOutsideAll level, pic, circle0, circle2, new_circle
        FindCircleOutsideAll level, pic, circle1, circle2, new_circle
    End If
End Sub

' Draw a circle tangent to these three circles and that is outside two of them.
Private Sub FindCircleOutsideTwo(ByVal level As Integer, ByVal pic As PictureBox, ByVal circle0 As MyCircle, ByVal circle1 As MyCircle, ByVal circle_contains As MyCircle)
Dim new_circle As MyCircle

    Set new_circle = FindApollonianCircle( _
        circle0, circle1, circle_contains, 1, 1, -1)
    If (new_circle Is Nothing) Then Exit Sub
    If (new_circle.Radius < 0.1) Then Exit Sub

    new_circle.Draw pic, vbRed

    level = level - 1
    If (level > 0) Then
        FindCircleOutsideTwo level, pic, new_circle, circle0, circle_contains
        FindCircleOutsideTwo level, pic, new_circle, circle1, circle_contains
        FindCircleOutsideAll level, pic, circle0, circle1, new_circle
    End If
End Sub

' Find a solution to Apollonius' problem.
' For discussion and method, see:
'    http://mathworld.wolfram.com/ApolloniusProblem.html
'    http://en.wikipedia.org/wiki/Problem_of_Apollonius#Algebraic_solutions
' For most of a Java code implementation, see:
'    http://www.diku.dk/hjemmesider/ansatte/rfonseca/implementations/apollonius.html
Private Function FindApollonianCircle(ByVal c1 As MyCircle, ByVal c2 As MyCircle, ByVal c3 As MyCircle, ByVal s1 As Integer, ByVal s2 As Integer, ByVal s3 As Integer) As MyCircle
Const tiny As Single = 0.0001

Dim temp_circle As MyCircle
Dim temp_s As Integer
Dim x1 As Single
Dim y1 As Single
Dim r1 As Single
Dim x2 As Single
Dim y2 As Single
Dim r2 As Single
Dim x3 As Single
Dim y3 As Single
Dim r3 As Single
Dim v11 As Single
Dim v12 As Single
Dim v13 As Single
Dim v14 As Single
Dim v21 As Single
Dim v22 As Single
Dim v23 As Single
Dim v24 As Single
Dim w12 As Single
Dim w13 As Single
Dim w14 As Single
Dim w22 As Single
Dim w23 As Single
Dim w24 As Single
Dim P As Single
Dim Q As Single
Dim M As Single
Dim N As Single
Dim a As Single
Dim b As Single
Dim c As Single
Dim solutions As Collection
Dim rs As Single
Dim xs As Single
Dim ys As Single
Dim new_circle As MyCircle

    ' Make sure c2 doesn't have the same X or Y coordinate as the others.
    If ((Abs(c2.X - c1.X) < tiny) Or (Abs(c2.Y - c1.Y) < tiny)) Then
        Set temp_circle = c2
        Set c2 = c3
        Set c3 = temp_circle
        temp_s = s2
        s2 = s3
        s3 = temp_s
    End If
    If ((Abs(c2.X - c3.X) < tiny) Or _
        (Abs(c2.Y - c3.Y) < tiny)) _
    Then
        Set temp_circle = c2
        Set c2 = c1
        Set c1 = temp_circle
        temp_s = s2
        s2 = s1
        s1 = temp_s
    End If

    x1 = c1.X
    y1 = c1.Y
    r1 = c1.Radius
    x2 = c2.X
    y2 = c2.Y
    r2 = c2.Radius
    x3 = c3.X
    y3 = c3.Y
    r3 = c3.Radius

    v11 = 2 * x2 - 2 * x1
    v12 = 2 * y2 - 2 * y1
    v13 = x1 * x1 - x2 * x2 + y1 * y1 - y2 * y2 - r1 * r1 + r2 * r2
    v14 = 2 * s2 * r2 - 2 * s1 * r1

    v21 = 2 * x3 - 2 * x2
    v22 = 2 * y3 - 2 * y2
    v23 = x2 * x2 - x3 * x3 + y2 * y2 - y3 * y3 - r2 * r2 + r3 * r3
    v24 = 2 * s3 * r3 - 2 * s2 * r2

    w12 = v12 / v11
    w13 = v13 / v11
    w14 = v14 / v11

    w22 = v22 / v21 - w12
    w23 = v23 / v21 - w13
    w24 = v24 / v21 - w14

    P = -w23 / w22
    Q = w24 / w22
    M = -w12 * P - w13
    N = w14 - w12 * Q

    a = N * N + Q * Q - 1
    b = 2 * M * N - 2 * N * x1 + 2 * P * Q - 2 * Q * y1 + 2 * s1 * r1
    c = x1 * x1 + M * M - 2 * M * x1 + P * P + y1 * y1 - 2 * P * y1 - r1 * r1

    ' Find roots of a quadratic equation
    Set solutions = QuadraticSolutions(a, b, c)
    If (solutions Is Nothing) Then
        Set FindApollonianCircle = Nothing
    Else
        rs = CSng(solutions(1))
        xs = M + N * rs
        ys = P + Q * rs
        Set new_circle = New MyCircle
        new_circle.Initialize xs, ys, rs
        Set FindApollonianCircle = new_circle
    End If
End Function

' Return solutions to a quadratic equation.
Private Function QuadraticSolutions(ByVal a As Double, ByVal b As Double, ByVal c As Double) As Collection
Const tiny As Double = 0.000001
Dim discriminant As Double
Dim results As Collection

    discriminant = b * b - 4 * a * c

    ' See if there are no real solutions.
    Set QuadraticSolutions = Nothing
    If (discriminant < 0) Then Exit Function

    ' See if there is one solution.
    Set results = New Collection
    If (discriminant < tiny) Then
        results.Add -b / (2 * a)
        Set QuadraticSolutions = results
        Exit Function
    End If

    ' There are two solutions.
    results.Add (-b + Sqr(discriminant)) / (2 * a)
    results.Add (-b - Sqr(discriminant)) / (2 * a)
    Set QuadraticSolutions = results
End Function
